from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_API_URL, PaginatedResults
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
from todoist_api_python.models import Label


@pytest.mark.asyncio
async def test_get_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
) -> None:
    label_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/labels/{label_id}"

    requests_mock.add(
        method=responses.GET,
        url=endpoint,
        json=default_label_response,
        status=200,
        match=[auth_matcher()],
    )

    label = todoist_api.get_label(label_id)

    assert len(requests_mock.calls) == 1
    assert label == default_label

    label = await todoist_api_async.get_label(label_id)

    assert len(requests_mock.calls) == 2
    assert label == default_label


@pytest.mark.asyncio
async def test_get_labels(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_labels_response: list[PaginatedResults],
    default_labels_list: list[list[Label]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/labels"

    cursor: str | None = None
    for page in default_labels_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher({}, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    labels_iter = todoist_api.get_labels()

    for i, labels in enumerate(labels_iter):
        assert len(requests_mock.calls) == count + 1
        assert labels == default_labels_list[i]
        count += 1

    labels_async_iter = await todoist_api_async.get_labels()

    async for i, labels in enumerate_async(labels_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert labels == default_labels_list[i]
        count += 1


@pytest.mark.asyncio
async def test_search_labels(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_labels_response: list[PaginatedResults],
    default_labels_list: list[list[Label]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/labels/search"
    query = "A label"

    cursor: str | None = None
    for page in default_labels_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"query": query}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    labels_iter = todoist_api.search_labels(query)

    for i, labels in enumerate(labels_iter):
        assert len(requests_mock.calls) == count + 1
        assert labels == default_labels_list[i]
        count += 1

    labels_async_iter = await todoist_api_async.search_labels(query)

    async for i, labels in enumerate_async(labels_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert labels == default_labels_list[i]
        count += 1


@pytest.mark.asyncio
async def test_add_label_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
) -> None:
    label_name = "A Label"

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/labels",
        json=default_label_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"name": label_name}),
        ],
    )

    new_label = todoist_api.add_label(name=label_name)

    assert len(requests_mock.calls) == 1
    assert new_label == default_label

    new_label = await todoist_api_async.add_label(name=label_name)

    assert len(requests_mock.calls) == 2
    assert new_label == default_label


@pytest.mark.asyncio
async def test_add_label_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
) -> None:
    label_name = "A Label"
    args: dict[str, Any] = {
        "color": "red",
        "item_order": 3,
        "is_favorite": True,
    }

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/labels",
        json=default_label_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"name": label_name} | args),
        ],
    )

    new_label = todoist_api.add_label(name=label_name, **args)

    assert len(requests_mock.calls) == 1
    assert new_label == default_label

    new_label = await todoist_api_async.add_label(name=label_name, **args)

    assert len(requests_mock.calls) == 2
    assert new_label == default_label


@pytest.mark.asyncio
async def test_update_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label: Label,
) -> None:
    args: dict[str, Any] = {
        "name": "An updated label",
    }
    updated_label_dict = default_label.to_dict() | args

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/labels/{default_label.id}",
        json=updated_label_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher(), data_matcher(args)],
    )

    response = todoist_api.update_label(label_id=default_label.id, **args)

    assert len(requests_mock.calls) == 1
    assert response == Label.from_dict(updated_label_dict)

    response = await todoist_api_async.update_label(label_id=default_label.id, **args)

    assert len(requests_mock.calls) == 2
    assert response == Label.from_dict(updated_label_dict)


@pytest.mark.asyncio
async def test_delete_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    label_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/labels/{label_id}"

    requests_mock.add(
        method=responses.DELETE,
        url=endpoint,
        status=204,
    )

    response = todoist_api.delete_label(label_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.delete_label(label_id)

    assert len(requests_mock.calls) == 2
    assert response is True



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, TypeVar, cast

from requests.status_codes import codes

from todoist_api_python._core.http_headers import create_headers

if TYPE_CHECKING:
    from requests import Session


# Timeouts for requests.
#
# 10 seconds for connecting is a recurring default and adheres to python-requests's
# recommendation of picking a value slightly larger than a multiple of 3.
#
# 60 seconds for reading aligns with Todoist's own internal timeout. All requests are
# forcefully terminated after this time, so there is no point waiting any longer.
TIMEOUT = (10, 60)

T = TypeVar("T")


def get(
    session: Session,
    url: str,
    token: str | None = None,
    request_id: str | None = None,
    params: dict[str, Any] | None = None,
) -> T:  # type: ignore[type-var]
    headers = create_headers(token=token, request_id=request_id)

    response = session.get(
        url,
        params=params,
        headers=headers,
        timeout=TIMEOUT,
    )

    if response.status_code == codes.OK:
        return cast("T", response.json())

    response.raise_for_status()
    return cast("T", response.ok)


def post(
    session: Session,
    url: str,
    token: str | None = None,
    request_id: str | None = None,
    *,
    params: dict[str, Any] | None = None,
    data: dict[str, Any] | None = None,
) -> T:  # type: ignore[type-var]
    headers = create_headers(
        token=token, with_content=bool(data), request_id=request_id
    )

    response = session.post(
        url,
        headers=headers,
        data=json.dumps(data) if data else None,
        params=params,
        timeout=TIMEOUT,
    )

    if response.status_code == codes.OK:
        return cast("T", response.json())

    response.raise_for_status()
    return cast("T", response.ok)


def delete(
    session: Session,
    url: str,
    token: str | None = None,
    request_id: str | None = None,
    params: dict[str, Any] | None = None,
) -> bool:
    headers = create_headers(token=token, request_id=request_id)

    response = session.delete(url, params=params, headers=headers, timeout=TIMEOUT)

    response.raise_for_status()
    return response.ok



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_API_URL, PaginatedResults
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
from todoist_api_python.models import Section


@pytest.mark.asyncio
async def test_get_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_section_response: dict[str, Any],
    default_section: Section,
) -> None:
    section_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/sections/{section_id}"

    requests_mock.add(
        method=responses.GET,
        url=endpoint,
        json=default_section_response,
        status=200,
        match=[auth_matcher()],
    )

    section = todoist_api.get_section(section_id)

    assert len(requests_mock.calls) == 1
    assert section == default_section

    section = await todoist_api_async.get_section(section_id)

    assert len(requests_mock.calls) == 2
    assert section == default_section


@pytest.mark.asyncio
async def test_get_sections(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[PaginatedResults],
    default_sections_list: list[list[Section]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/sections"

    cursor: str | None = None
    for page in default_sections_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher({}, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    sections_iter = todoist_api.get_sections()

    for i, sections in enumerate(sections_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1

    sections_async_iter = await todoist_api_async.get_sections()

    async for i, sections in enumerate_async(sections_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1


@pytest.mark.asyncio
async def test_get_sections_by_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[PaginatedResults],
    default_sections_list: list[list[Section]],
) -> None:
    project_id = "123"
    endpoint = f"{DEFAULT_API_URL}/sections"

    cursor: str | None = None
    for page in default_sections_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"project_id": project_id}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    sections_iter = todoist_api.get_sections(project_id=project_id)

    for i, sections in enumerate(sections_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1

    sections_async_iter = await todoist_api_async.get_sections(project_id=project_id)

    async for i, sections in enumerate_async(sections_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1


@pytest.mark.asyncio
async def test_search_sections(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[PaginatedResults],
    default_sections_list: list[list[Section]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/sections/search"
    query = "A Section"

    cursor: str | None = None
    for page in default_sections_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"query": query}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    sections_iter = todoist_api.search_sections(query)

    for i, sections in enumerate(sections_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1

    sections_async_iter = await todoist_api_async.search_sections(query)

    async for i, sections in enumerate_async(sections_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1


@pytest.mark.asyncio
async def test_search_sections_by_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[PaginatedResults],
    default_sections_list: list[list[Section]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/sections/search"
    project_id = "123"
    query = "A Section"

    cursor: str | None = None
    for page in default_sections_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"query": query, "project_id": project_id}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    sections_iter = todoist_api.search_sections(query, project_id=project_id)

    for i, sections in enumerate(sections_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1

    sections_async_iter = await todoist_api_async.search_sections(
        query, project_id=project_id
    )

    async for i, sections in enumerate_async(sections_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert sections == default_sections_list[i]
        count += 1


@pytest.mark.asyncio
async def test_add_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_section_response: dict[str, Any],
    default_section: Section,
) -> None:
    section_name = "A Section"
    project_id = "123"
    args = {
        "order": 3,
    }

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/sections",
        json=default_section_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"name": section_name, "project_id": project_id} | args),
        ],
    )

    new_section = todoist_api.add_section(
        name=section_name, project_id=project_id, **args
    )

    assert len(requests_mock.calls) == 1
    assert new_section == default_section

    new_section = await todoist_api_async.add_section(
        name=section_name, project_id=project_id, **args
    )

    assert len(requests_mock.calls) == 2
    assert new_section == default_section


@pytest.mark.asyncio
async def test_update_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_section: Section,
) -> None:
    args = {
        "name": "An updated section",
    }
    updated_section_dict = default_section.to_dict() | args

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/sections/{default_section.id}",
        json=updated_section_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher(), data_matcher(args)],
    )

    response = todoist_api.update_section(section_id=default_section.id, **args)

    assert len(requests_mock.calls) == 1
    assert response == Section.from_dict(updated_section_dict)

    response = await todoist_api_async.update_section(
        section_id=default_section.id, **args
    )

    assert len(requests_mock.calls) == 2
    assert response == Section.from_dict(updated_section_dict)


@pytest.mark.asyncio
async def test_delete_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    section_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/sections/{section_id}"

    requests_mock.add(
        method=responses.DELETE,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.delete_section(section_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.delete_section(section_id)

    assert len(requests_mock.calls) == 2
    assert response is True



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_API_URL, PaginatedResults
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
from todoist_api_python.models import Collaborator, Project


@pytest.mark.asyncio
async def test_get_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project_response: dict[str, Any],
    default_project: Project,
) -> None:
    project_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/projects/{project_id}"

    requests_mock.add(
        method=responses.GET,
        url=endpoint,
        json=default_project_response,
        status=200,
        match=[auth_matcher()],
    )

    project = todoist_api.get_project(project_id)

    assert len(requests_mock.calls) == 1
    assert project == default_project

    project = await todoist_api_async.get_project(project_id)

    assert len(requests_mock.calls) == 2
    assert project == default_project


@pytest.mark.asyncio
async def test_get_projects(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_projects_response: list[PaginatedResults],
    default_projects_list: list[list[Project]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/projects"

    cursor: str | None = None
    for page in default_projects_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher({}, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    projects_iter = todoist_api.get_projects()

    for i, projects in enumerate(projects_iter):
        assert len(requests_mock.calls) == count + 1
        assert projects == default_projects_list[i]
        count += 1

    projects_async_iter = await todoist_api_async.get_projects()

    async for i, projects in enumerate_async(projects_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert projects == default_projects_list[i]
        count += 1


@pytest.mark.asyncio
async def test_search_projects(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_projects_response: list[PaginatedResults],
    default_projects_list: list[list[Project]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/projects/search"
    query = "Inbox"

    cursor: str | None = None
    for page in default_projects_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"query": query}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    projects_iter = todoist_api.search_projects(query)

    for i, projects in enumerate(projects_iter):
        assert len(requests_mock.calls) == count + 1
        assert projects == default_projects_list[i]
        count += 1

    projects_async_iter = await todoist_api_async.search_projects(query)

    async for i, projects in enumerate_async(projects_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert projects == default_projects_list[i]
        count += 1


@pytest.mark.asyncio
async def test_add_project_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project_response: dict[str, Any],
    default_project: Project,
) -> None:
    project_name = "A Project"

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/projects",
        json=default_project_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"name": project_name}),
        ],
    )

    new_project = todoist_api.add_project(name=project_name)

    assert len(requests_mock.calls) == 1
    assert new_project == default_project

    new_project = await todoist_api_async.add_project(name=project_name)

    assert len(requests_mock.calls) == 2
    assert new_project == default_project


@pytest.mark.asyncio
async def test_add_project_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project_response: dict[str, Any],
    default_project: Project,
) -> None:
    project_name = "A Project"

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/projects",
        json=default_project_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"name": project_name}),
        ],
    )

    new_project = todoist_api.add_project(name=project_name)

    assert len(requests_mock.calls) == 1
    assert new_project == default_project

    new_project = await todoist_api_async.add_project(name=project_name)

    assert len(requests_mock.calls) == 2
    assert new_project == default_project


@pytest.mark.asyncio
async def test_update_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project: Project,
) -> None:
    args: dict[str, Any] = {
        "name": "An updated project",
        "color": "red",
        "is_favorite": False,
    }
    updated_project_dict = default_project.to_dict() | args

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/projects/{default_project.id}",
        json=updated_project_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher(), data_matcher(args)],
    )

    response = todoist_api.update_project(project_id=default_project.id, **args)

    assert len(requests_mock.calls) == 1
    assert response == Project.from_dict(updated_project_dict)

    response = await todoist_api_async.update_project(
        project_id=default_project.id, **args
    )

    assert len(requests_mock.calls) == 2
    assert response == Project.from_dict(updated_project_dict)


@pytest.mark.asyncio
async def test_archive_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project: Project,
) -> None:
    project_id = default_project.id
    endpoint = f"{DEFAULT_API_URL}/projects/{project_id}/archive"

    archived_project_dict = default_project.to_dict()
    archived_project_dict["is_archived"] = True

    requests_mock.add(
        method=responses.POST,
        url=endpoint,
        json=archived_project_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher()],
    )

    project = todoist_api.archive_project(project_id)

    assert len(requests_mock.calls) == 1
    assert project == Project.from_dict(archived_project_dict)

    project = await todoist_api_async.archive_project(project_id)

    assert len(requests_mock.calls) == 2
    assert project == Project.from_dict(archived_project_dict)


@pytest.mark.asyncio
async def test_unarchive_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_project: Project,
) -> None:
    project_id = default_project.id
    endpoint = f"{DEFAULT_API_URL}/projects/{project_id}/unarchive"

    unarchived_project_dict = default_project.to_dict()
    unarchived_project_dict["is_archived"] = False

    requests_mock.add(
        method=responses.POST,
        url=endpoint,
        json=unarchived_project_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher()],
    )

    project = todoist_api.unarchive_project(project_id)

    assert len(requests_mock.calls) == 1
    assert project == Project.from_dict(unarchived_project_dict)

    project = await todoist_api_async.unarchive_project(project_id)

    assert len(requests_mock.calls) == 2
    assert project == Project.from_dict(unarchived_project_dict)


@pytest.mark.asyncio
async def test_delete_project(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    project_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/projects/{project_id}"

    requests_mock.add(
        method=responses.DELETE,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.delete_project(project_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.delete_project(project_id)

    assert len(requests_mock.calls) == 2
    assert response is True


@pytest.mark.asyncio
async def test_get_collaborators(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_collaborators_response: list[PaginatedResults],
    default_collaborators_list: list[list[Collaborator]],
) -> None:
    project_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/projects/{project_id}/collaborators"

    cursor: str | None = None
    for page in default_collaborators_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher({}, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    collaborators_iter = todoist_api.get_collaborators(project_id)

    for i, collaborators in enumerate(collaborators_iter):
        assert len(requests_mock.calls) == count + 1
        assert collaborators == default_collaborators_list[i]
        count += 1

    collaborators_async_iter = await todoist_api_async.get_collaborators(project_id)

    async for i, collaborators in enumerate_async(collaborators_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert collaborators == default_collaborators_list[i]
        count += 1



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import (
    DEFAULT_API_URL,
    PaginatedResults,
)
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)
from todoist_api_python.models import Attachment

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync

from todoist_api_python.models import Comment


@pytest.mark.asyncio
async def test_get_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comment_response: dict[str, Any],
    default_comment: Comment,
) -> None:
    comment_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/comments/{comment_id}"

    requests_mock.add(
        method=responses.GET,
        url=endpoint,
        json=default_comment_response,
        status=200,
        match=[auth_matcher(), request_id_matcher()],
    )

    comment = todoist_api.get_comment(comment_id)

    assert len(requests_mock.calls) == 1
    assert comment == default_comment

    comment = await todoist_api_async.get_comment(comment_id)

    assert len(requests_mock.calls) == 2
    assert comment == default_comment


@pytest.mark.asyncio
async def test_get_comments(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comments_response: list[PaginatedResults],
    default_comments_list: list[list[Comment]],
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/comments"

    cursor: str | None = None
    for page in default_comments_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[
                auth_matcher(),
                request_id_matcher(),
                param_matcher({"task_id": task_id}, cursor),
            ],
        )
        cursor = page["next_cursor"]

    count = 0

    comments_iter = todoist_api.get_comments(task_id=task_id)

    for i, comments in enumerate(comments_iter):
        assert len(requests_mock.calls) == count + 1
        assert comments == default_comments_list[i]
        count += 1

    comments_async_iter = await todoist_api_async.get_comments(task_id=task_id)

    async for i, comments in enumerate_async(comments_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert comments == default_comments_list[i]
        count += 1


@pytest.mark.asyncio
async def test_add_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comment_response: dict[str, Any],
    default_comment: Comment,
) -> None:
    content = "A Comment"
    project_id = "6HWcc9PJCvPjCxC9"
    attachment = Attachment(
        resource_type="file",
        file_url="https://s3.amazonaws.com/domorebetter/Todoist+Setup+Guide.pdf",
        file_type="application/pdf",
        file_name="File.pdf",
    )

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/comments",
        json=default_comment_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher(
                {
                    "content": content,
                    "project_id": project_id,
                    "attachment": attachment.to_dict(),
                }
            ),
        ],
    )

    new_comment = todoist_api.add_comment(
        content=content,
        project_id=project_id,
        attachment=attachment,
    )

    assert len(requests_mock.calls) == 1
    assert new_comment == default_comment

    new_comment = await todoist_api_async.add_comment(
        content=content,
        project_id=project_id,
        attachment=attachment,
    )

    assert len(requests_mock.calls) == 2
    assert new_comment == default_comment


@pytest.mark.asyncio
async def test_update_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comment: Comment,
) -> None:
    args = {
        "content": "An updated comment",
    }
    updated_comment_dict = default_comment.to_dict() | args

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/comments/{default_comment.id}",
        json=updated_comment_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher(), data_matcher(args)],
    )

    response = todoist_api.update_comment(comment_id=default_comment.id, **args)

    assert len(requests_mock.calls) == 1
    assert response == Comment.from_dict(updated_comment_dict)

    response = await todoist_api_async.update_comment(
        comment_id=default_comment.id, **args
    )

    assert len(requests_mock.calls) == 2
    assert response == Comment.from_dict(updated_comment_dict)


@pytest.mark.asyncio
async def test_delete_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    comment_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/comments/{comment_id}"

    requests_mock.add(
        method=responses.DELETE,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.delete_comment(comment_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.delete_comment(comment_id)

    assert len(requests_mock.calls) == 2
    assert response is True



from __future__ import annotations

import sys
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    UTC = timezone.utc

import pytest
import responses

from tests.data.test_defaults import DEFAULT_API_URL, PaginatedResults
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
from todoist_api_python.models import Task


@pytest.mark.asyncio
async def test_get_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/tasks/{task_id}"

    requests_mock.add(
        method=responses.GET,
        url=endpoint,
        json=default_task_response,
        match=[auth_matcher(), request_id_matcher()],
    )

    task = todoist_api.get_task(task_id)

    assert len(requests_mock.calls) == 1
    assert task == default_task

    task = await todoist_api_async.get_task(task_id)

    assert len(requests_mock.calls) == 2
    assert task == default_task


@pytest.mark.asyncio
async def test_get_tasks(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_tasks_response: list[PaginatedResults],
    default_tasks_list: list[list[Task]],
) -> None:
    endpoint = f"{DEFAULT_API_URL}/tasks"

    cursor: str | None = None
    for page in default_tasks_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher({}, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    tasks_iter = todoist_api.get_tasks()

    for i, tasks in enumerate(tasks_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1

    tasks_async_iter = await todoist_api_async.get_tasks()

    async for i, tasks in enumerate_async(tasks_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1


@pytest.mark.asyncio
async def test_get_tasks_with_filters(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_tasks_response: list[PaginatedResults],
    default_tasks_list: list[list[Task]],
) -> None:
    project_id = "123"
    section_id = "456"
    parent_id = "789"
    label = "test-label"
    ids = ["1", "2", "3"]
    limit = 30

    params: dict[str, Any] = {
        "project_id": project_id,
        "section_id": section_id,
        "parent_id": parent_id,
        "label": label,
        "ids": ",".join(ids),
        "limit": limit,
    }

    endpoint = f"{DEFAULT_API_URL}/tasks"

    cursor: str | None = None
    for page in default_tasks_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher(params, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    tasks_iter = todoist_api.get_tasks(
        project_id=project_id,
        section_id=section_id,
        parent_id=parent_id,
        label=label,
        ids=ids,
        limit=limit,
    )

    for i, tasks in enumerate(tasks_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1

    tasks_async_iter = await todoist_api_async.get_tasks(
        project_id=project_id,
        section_id=section_id,
        parent_id=parent_id,
        label=label,
        ids=ids,
        limit=limit,
    )

    async for i, tasks in enumerate_async(tasks_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1


@pytest.mark.asyncio
async def test_filter_tasks(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_tasks_response: list[PaginatedResults],
    default_tasks_list: list[list[Task]],
) -> None:
    query = "today or overdue"
    lang = "en"
    params = {
        "query": "today or overdue",
        "lang": "en",
    }

    endpoint = f"{DEFAULT_API_URL}/tasks/filter"

    cursor: str | None = None
    for page in default_tasks_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher(params, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    tasks_iter = todoist_api.filter_tasks(
        query=query,
        lang=lang,
    )

    for i, tasks in enumerate(tasks_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1

    # Test async iterator
    tasks_async_iter = await todoist_api_async.filter_tasks(
        query=query,
        lang=lang,
    )

    async for i, tasks in enumerate_async(tasks_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_tasks_list[i]
        count += 1


@pytest.mark.asyncio
async def test_add_task_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
) -> None:
    content = "Some content"

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/tasks",
        json=default_task_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher({"content": content}),
        ],
    )

    new_task = todoist_api.add_task(content=content)

    assert len(requests_mock.calls) == 1
    assert new_task == default_task

    new_task = await todoist_api_async.add_task(content=content)

    assert len(requests_mock.calls) == 2
    assert new_task == default_task


@pytest.mark.asyncio
async def test_add_task_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
) -> None:
    content = "Some content"
    due_datetime = datetime(2021, 1, 1, 11, 0, 0, tzinfo=UTC)
    args: dict[str, Any] = {
        "description": "A description",
        "project_id": "123",
        "section_id": "456",
        "parent_id": "789",
        "labels": ["label1", "label2"],
        "priority": 4,
        "due_string": "today",
        "due_lang": "en",
        "assignee_id": "321",
        "order": 3,
        "auto_reminder": True,
        "auto_parse_labels": True,
        "duration": 60,
        "duration_unit": "minute",
    }

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/tasks",
        json=default_task_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher(
                {
                    "content": content,
                    "due_datetime": due_datetime.strftime("%Y-%m-%dT%H:%M:%SZ"),
                }
                | args
            ),
        ],
    )

    new_task = todoist_api.add_task(content=content, due_datetime=due_datetime, **args)

    assert len(requests_mock.calls) == 1
    assert new_task == default_task

    new_task = await todoist_api_async.add_task(
        content=content, due_datetime=due_datetime, **args
    )

    assert len(requests_mock.calls) == 2
    assert new_task == default_task


@pytest.mark.asyncio
async def test_add_task_quick(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_meta_response: dict[str, Any],
    default_task_meta: Task,
) -> None:
    text = "Buy milk tomorrow at 9am #Shopping @errands"
    note = "Whole milk x6"
    auto_reminder = True

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/tasks/quick",
        json=default_task_meta_response,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(),
            data_matcher(
                {
                    "meta": True,
                    "text": text,
                    "auto_reminder": auto_reminder,
                    "note": note,
                }
            ),
        ],
    )

    task = todoist_api.add_task_quick(
        text=text,
        note=note,
        auto_reminder=auto_reminder,
    )

    assert len(requests_mock.calls) == 1
    assert task == default_task_meta

    task = await todoist_api_async.add_task_quick(
        text=text,
        note=note,
        auto_reminder=auto_reminder,
    )

    assert len(requests_mock.calls) == 2
    assert task == default_task_meta


@pytest.mark.asyncio
async def test_update_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task: Task,
) -> None:
    args: dict[str, Any] = {
        "content": "Updated content",
        "description": "Updated description",
        "labels": ["label1", "label2"],
        "priority": 2,
    }
    updated_task_dict = default_task.to_dict() | args

    requests_mock.add(
        method=responses.POST,
        url=f"{DEFAULT_API_URL}/tasks/{default_task.id}",
        json=updated_task_dict,
        status=200,
        match=[auth_matcher(), request_id_matcher(), data_matcher(args)],
    )

    response = todoist_api.update_task(task_id=default_task.id, **args)

    assert len(requests_mock.calls) == 1
    assert response == Task.from_dict(updated_task_dict)

    response = await todoist_api_async.update_task(task_id=default_task.id, **args)

    assert len(requests_mock.calls) == 2
    assert response == Task.from_dict(updated_task_dict)


@pytest.mark.asyncio
async def test_complete_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/tasks/{task_id}/close"

    requests_mock.add(
        method=responses.POST,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.complete_task(task_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.complete_task(task_id)

    assert len(requests_mock.calls) == 2
    assert response is True


@pytest.mark.asyncio
async def test_uncomplete_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/tasks/{task_id}/reopen"

    requests_mock.add(
        method=responses.POST,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.uncomplete_task(task_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.uncomplete_task(task_id)

    assert len(requests_mock.calls) == 2
    assert response is True


@pytest.mark.asyncio
async def test_move_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/tasks/{task_id}/move"

    requests_mock.add(
        method=responses.POST,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.move_task(task_id, project_id="123")

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.move_task(task_id, section_id="456")

    assert len(requests_mock.calls) == 2
    assert response is True

    response = await todoist_api_async.move_task(task_id, parent_id="789")

    assert len(requests_mock.calls) == 3
    assert response is True

    with pytest.raises(
        ValueError,
        match="Either `project_id`, `section_id`, or `parent_id` must be provided.",
    ):
        response = await todoist_api_async.move_task(task_id)


@pytest.mark.asyncio
async def test_delete_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
) -> None:
    task_id = "6X7rM8997g3RQmvh"
    endpoint = f"{DEFAULT_API_URL}/tasks/{task_id}"

    requests_mock.add(
        method=responses.DELETE,
        url=endpoint,
        status=204,
        match=[auth_matcher(), request_id_matcher()],
    )

    response = todoist_api.delete_task(task_id)

    assert len(requests_mock.calls) == 1
    assert response is True

    response = await todoist_api_async.delete_task(task_id)

    assert len(requests_mock.calls) == 2
    assert response is True



from __future__ import annotations

from typing import Any

import pytest
import responses
from requests import HTTPError, Session
from responses.matchers import query_param_matcher

from tests.data.test_defaults import DEFAULT_REQUEST_ID, DEFAULT_TOKEN
from tests.utils.test_utils import (
    auth_matcher,
    data_matcher,
    param_matcher,
    request_id_matcher,
)
from todoist_api_python._core.http_requests import delete, get, post

EXAMPLE_URL = "https://example.com/"
EXAMPLE_PARAMS = {"param1": "value1", "param2": "value2"}
EXAMPLE_DATA = {"param3": "value31", "param4": "value4"}
EXAMPLE_RESPONSE = {"result": "ok"}


@responses.activate
def test_get_with_params(default_task_response: dict[str, Any]) -> None:
    responses.add(
        method=responses.GET,
        url=EXAMPLE_URL,
        json=EXAMPLE_RESPONSE,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(DEFAULT_REQUEST_ID),
            param_matcher(EXAMPLE_PARAMS),
        ],
    )

    response: dict[str, Any] = get(
        session=Session(),
        url=EXAMPLE_URL,
        token=DEFAULT_TOKEN,
        request_id=DEFAULT_REQUEST_ID,
        params=EXAMPLE_PARAMS,
    )

    assert len(responses.calls) == 1
    assert response == EXAMPLE_RESPONSE


@responses.activate
def test_get_raise_for_status() -> None:
    responses.add(
        method=responses.GET,
        url=EXAMPLE_URL,
        json="<error description>",
        status=500,
    )

    with pytest.raises(HTTPError) as error_info:
        get(Session(), EXAMPLE_URL, DEFAULT_TOKEN)

    assert error_info.value.response.content == b'"<error description>"'


@responses.activate
def test_post_with_data(default_task_response: dict[str, Any]) -> None:
    responses.add(
        method=responses.POST,
        url=EXAMPLE_URL,
        json=EXAMPLE_RESPONSE,
        status=200,
        match=[
            auth_matcher(),
            request_id_matcher(DEFAULT_REQUEST_ID),
            data_matcher(EXAMPLE_DATA),
        ],
    )

    response: dict[str, Any] = post(
        session=Session(),
        url=EXAMPLE_URL,
        token=DEFAULT_TOKEN,
        request_id=DEFAULT_REQUEST_ID,
        data=EXAMPLE_DATA,
    )

    assert len(responses.calls) == 1
    assert response == EXAMPLE_RESPONSE


@responses.activate
def test_post_return_ok_when_no_response_body() -> None:
    responses.add(
        method=responses.POST,
        url=EXAMPLE_URL,
        status=204,
    )

    result: bool = post(session=Session(), url=EXAMPLE_URL, token=DEFAULT_TOKEN)
    assert result is True


@responses.activate
def test_post_raise_for_status() -> None:
    responses.add(
        method=responses.POST,
        url=EXAMPLE_URL,
        status=500,
    )

    with pytest.raises(HTTPError):
        post(session=Session(), url=EXAMPLE_URL, token=DEFAULT_TOKEN)


@responses.activate
def test_delete_with_params() -> None:
    responses.add(
        method=responses.DELETE,
        url=EXAMPLE_URL,
        status=204,
        match=[
            auth_matcher(),
            request_id_matcher(DEFAULT_REQUEST_ID),
            query_param_matcher(EXAMPLE_PARAMS),
        ],
    )

    result = delete(
        session=Session(),
        url=EXAMPLE_URL,
        token=DEFAULT_TOKEN,
        request_id=DEFAULT_REQUEST_ID,
        params=EXAMPLE_PARAMS,
    )

    assert len(responses.calls) == 1
    assert result is True


@responses.activate
def test_delete_raise_for_status() -> None:
    responses.add(
        method=responses.DELETE,
        url=EXAMPLE_URL,
        status=500,
    )

    with pytest.raises(HTTPError):
        delete(session=Session(), url=EXAMPLE_URL, token=DEFAULT_TOKEN)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, TypeVar

from responses import matchers

from tests.data.test_defaults import (
    DEFAULT_TOKEN,
)
from todoist_api_python.api import TodoistAPI

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, AsyncIterator, Callable


RE_UUID = re.compile(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", re.IGNORECASE)


def auth_matcher() -> Callable[..., Any]:
    return matchers.header_matcher({"Authorization": f"Bearer {DEFAULT_TOKEN}"})


def request_id_matcher(request_id: str | None = None) -> Callable[..., Any]:
    return matchers.header_matcher({"X-Request-Id": request_id or RE_UUID})


def param_matcher(
    params: dict[str, str], cursor: str | None = None
) -> Callable[..., Any]:
    return matchers.query_param_matcher(params | ({"cursor": cursor} if cursor else {}))


def data_matcher(data: dict[str, Any]) -> Callable[..., Any]:
    return matchers.json_params_matcher(data)


def get_todoist_api_patch(method: Callable[..., Any] | None) -> str:
    module = TodoistAPI.__module__
    name = TodoistAPI.__qualname__

    return f"{module}.{name}.{method.__name__}" if method else f"{module}.{name}"


T = TypeVar("T")


async def enumerate_async(
    iterable: AsyncIterable[T], start: int = 0
) -> AsyncIterator[tuple[int, T]]:
    index = start
    async for value in iterable:
        yield index, value
        index += 1



from __future__ import annotations

import asyncio
import sys
import uuid
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, TypeVar, cast

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Iterator

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    UTC = timezone.utc

T = TypeVar("T")


async def run_async(func: Callable[[], T]) -> T:
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, func)


async def generate_async(iterator: Iterator[T]) -> AsyncGenerator[T]:
    def get_next_item() -> tuple[bool, T | None]:
        try:
            return True, next(iterator)
        except StopIteration:
            return False, None

    while True:
        has_more, item = await run_async(get_next_item)
        if has_more is True:
            yield cast("T", item)
        else:
            break


def format_date(d: date) -> str:
    """Format a date object as YYYY-MM-DD."""
    return d.isoformat()


def format_datetime(dt: datetime) -> str:
    """
    Format a datetime object.

    YYYY-MM-DDTHH:MM:SS for naive datetimes; YYYY-MM-DDTHH:MM:SSZ for aware datetimes.
    """
    if dt.tzinfo is None:
        return dt.isoformat()
    return dt.astimezone(UTC).isoformat().replace("+00:00", "Z")


def parse_date(date_str: str) -> date:
    """Parse a YYYY-MM-DD string into a date object."""
    return date.fromisoformat(date_str)


def parse_datetime(datetime_str: str) -> datetime:
    """
    Parse a string into a datetime object.

    YYYY-MM-DDTHH:MM:SS for naive datetimes; YYYY-MM-DDTHH:MM:SSZ for aware datetimes.
    """
    from datetime import datetime

    if datetime_str.endswith("Z"):
        datetime_str = datetime_str[:-1] + "+00:00"
        return datetime.fromisoformat(datetime_str)
    return datetime.fromisoformat(datetime_str)


def default_request_id_fn() -> str:
    """Generate random UUIDv4s as the default request ID."""
    return str(uuid.uuid4())



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Literal, Optional, Union

from dataclass_wizard import JSONPyWizard
from dataclass_wizard.v1 import DatePattern, DateTimePattern, UTCDateTimePattern
from dataclass_wizard.v1.models import Alias

from todoist_api_python._core.endpoints import INBOX_URL, get_project_url, get_task_url

ViewStyle = Literal["list", "board", "calendar"]
DurationUnit = Literal["minute", "day"]
ApiDate = UTCDateTimePattern["%FT%T.%fZ"]  # type: ignore[valid-type]
ApiDue = Union[  # https://github.com/rnag/dataclass-wizard/issues/189
    DatePattern["%F"], DateTimePattern["%FT%T"], UTCDateTimePattern["%FT%TZ"]  # type: ignore[valid-type]  # noqa: F722
]


@dataclass
class Project(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    name: str
    description: str
    order: Annotated[int, Alias(load=("child_order", "order"))]
    color: str
    is_collapsed: Annotated[bool, Alias(load=("collapsed", "is_collapsed"))]
    is_shared: Annotated[bool, Alias(load=("shared", "is_shared"))]
    is_favorite: bool
    is_archived: bool
    can_assign_tasks: bool
    view_style: ViewStyle
    created_at: ApiDate
    updated_at: ApiDate

    parent_id: str | None = None
    is_inbox_project: Annotated[
        bool | None, Alias(load=("inbox_project", "is_inbox_project"))
    ] = None

    workspace_id: str | None = None
    folder_id: str | None = None

    @property
    def url(self) -> str:
        if self.is_inbox_project:
            return INBOX_URL
        return get_project_url(self.id, self.name)


@dataclass
class Section(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    name: str
    project_id: str
    is_collapsed: Annotated[bool, Alias(load=("collapsed", "is_collapsed"))]
    order: Annotated[int, Alias(load=("section_order", "order"))]


@dataclass
class Due(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    date: ApiDue
    string: str
    lang: str = "en"
    is_recurring: bool = False
    timezone: str | None = None


@dataclass
class Deadline(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    date: ApiDue
    lang: str = "en"


@dataclass
class Meta(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    project: tuple[str | None, str | None]
    section: tuple[str | None, str | None]
    assignee: tuple[str | None, str | None]
    labels: dict[int, str]
    due: Due | None
    deadline: Deadline | None


@dataclass
class Task(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    content: str
    description: str
    project_id: str
    section_id: str | None
    parent_id: str | None
    labels: list[str] | None
    priority: int
    due: Due | None
    deadline: Deadline | None
    duration: Duration | None
    is_collapsed: Annotated[bool, Alias(load=("collapsed", "is_collapsed"))]
    order: Annotated[int, Alias(load=("child_order", "order"))]
    assignee_id: Annotated[str | None, Alias(load=("responsible_uid", "assignee_id"))]
    assigner_id: Annotated[str | None, Alias(load=("assigned_by_uid", "assigner_id"))]
    completed_at: Optional[ApiDate]  # noqa: UP007 # https://github.com/rnag/dataclass-wizard/issues/189
    creator_id: Annotated[str, Alias(load=("added_by_uid", "creator_id"))]
    created_at: Annotated[ApiDate, Alias(load=("added_at", "created_at"))]
    updated_at: ApiDate

    meta: Meta | None = None

    @property
    def url(self) -> str:
        return get_task_url(self.id, self.content)

    @property
    def is_completed(self) -> bool:
        return self.completed_at is not None


@dataclass
class Collaborator(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    email: str
    name: str


@dataclass
class Attachment(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    resource_type: str | None = None

    file_name: str | None = None
    file_size: int | None = None
    file_type: str | None = None
    file_url: str | None = None
    file_duration: int | None = None
    upload_state: str | None = None

    image: str | None = None
    image_width: int | None = None
    image_height: int | None = None

    url: str | None = None
    title: str | None = None


@dataclass
class Comment(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    content: str
    poster_id: Annotated[str, Alias(load=("posted_uid", "poster_id"))]
    posted_at: ApiDate
    task_id: Annotated[str | None, Alias(load=("item_id", "task_id"))] = None
    project_id: str | None = None
    attachment: Annotated[
        Attachment | None, Alias(load=("file_attachment", "attachment"))
    ] = None

    def __post_init__(self) -> None:
        """
        Finish initialization of the Comment object.

        :raises ValueError: If neither `task_id` nor `project_id` is specified.
        """
        if self.task_id is None and self.project_id is None:
            raise ValueError("Must specify `task_id` or `project_id`")


@dataclass
class Label(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    id: str
    name: str
    color: str
    order: int
    is_favorite: bool


@dataclass
class AuthResult(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    access_token: str
    state: str | None


@dataclass
class Duration(JSONPyWizard):
    class _(JSONPyWizard.Meta):  # noqa:N801
        v1 = True

    amount: int
    unit: DurationUnit



from __future__ import annotations

import sys
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeVar
from weakref import finalize

import requests
from annotated_types import Ge, Le, MaxLen, MinLen, Predicate

from todoist_api_python._core.endpoints import (
    COLLABORATORS_PATH,
    COMMENTS_PATH,
    LABELS_PATH,
    LABELS_SEARCH_PATH_SUFFIX,
    PROJECT_ARCHIVE_PATH_SUFFIX,
    PROJECT_UNARCHIVE_PATH_SUFFIX,
    PROJECTS_PATH,
    PROJECTS_SEARCH_PATH_SUFFIX,
    SECTIONS_PATH,
    SECTIONS_SEARCH_PATH_SUFFIX,
    SHARED_LABELS_PATH,
    SHARED_LABELS_REMOVE_PATH,
    SHARED_LABELS_RENAME_PATH,
    TASKS_COMPLETED_BY_COMPLETION_DATE_PATH,
    TASKS_COMPLETED_BY_DUE_DATE_PATH,
    TASKS_FILTER_PATH,
    TASKS_PATH,
    TASKS_QUICK_ADD_PATH,
    get_api_url,
)
from todoist_api_python._core.http_requests import delete, get, post
from todoist_api_python._core.utils import (
    default_request_id_fn,
    format_date,
    format_datetime,
)
from todoist_api_python.models import (
    Attachment,
    Collaborator,
    Comment,
    Label,
    Project,
    Section,
    Task,
)

if TYPE_CHECKING:
    from datetime import date, datetime
    from types import TracebackType

if sys.version_info >= (3, 11):
    from typing import Self
else:
    Self = TypeVar("Self", bound="TodoistAPI")


LanguageCode = Annotated[str, Predicate(lambda x: len(x) == 2)]  # noqa: PLR2004
ColorString = Annotated[
    str,
    Predicate(
        lambda x: x
        in (
            "berry_red",
            "red",
            "orange",
            "yellow",
            "olive_green",
            "lime_green",
            "green",
            "mint_green",
            "teal",
            "sky_blue",
            "light_blue",
            "blue",
            "grape",
            "violet",
            "lavender",
            "magenta",
            "salmon",
            "charcoal",
            "grey",
            "taupe",
        )
    ),
]
ViewStyle = Annotated[str, Predicate(lambda x: x in ("list", "board", "calendar"))]


class TodoistAPI:
    """
    Client for the Todoist API.

    Provides methods for interacting with Todoist resources like tasks, projects,
    labels, comments, etc.

    Manages an HTTP session and handles authentication. Can be used as a context manager
    to ensure the session is closed properly.
    """

    def __init__(
        self,
        token: str,
        request_id_fn: Callable[[], str] | None = default_request_id_fn,
        session: requests.Session | None = None,
    ) -> None:
        """
        Initialize the TodoistAPI client.

        :param token: Authentication token for the Todoist API.
        :param request_id_fn: Generator of request IDs for the `X-Request-ID` header.
        :param session: An optional pre-configured requests `Session` object.
        """
        self._token = token
        self._request_id_fn = request_id_fn
        self._session = session or requests.Session()
        self._finalizer = finalize(self, self._session.close)

    def __enter__(self) -> Self:
        """
        Enters the runtime context related to this object.

        The with statement will bind this method's return value to the target(s)
        specified in the as clause of the statement, if any.

        :return: This TodoistAPI instance.
        """
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the runtime context and closes the underlying requests session."""
        self._finalizer()

    def get_task(self, task_id: str) -> Task:
        """
        Get a specific task by its ID.

        :param task_id: The ID of the task to retrieve.
        :return: The requested task.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Task dictionary.
        """
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}")
        task_data: dict[str, Any] = get(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Task.from_dict(task_data)

    def get_tasks(
        self,
        *,
        project_id: str | None = None,
        section_id: str | None = None,
        parent_id: str | None = None,
        label: str | None = None,
        ids: list[str] | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Task]]:
        """
        Get an iterable of lists of active tasks.

        The response is an iterable of lists of active tasks matching the criteria.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param project_id: Filter tasks by project ID.
        :param section_id: Filter tasks by section ID.
        :param parent_id: Filter tasks by parent task ID.
        :param label: Filter tasks by label name.
        :param ids: A list of the IDs of the tasks to retrieve.
        :param limit: Maximum number of tasks per page.
        :return: An iterable of lists of tasks.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(TASKS_PATH)

        params: dict[str, Any] = {}
        if project_id is not None:
            params["project_id"] = project_id
        if section_id is not None:
            params["section_id"] = section_id
        if parent_id is not None:
            params["parent_id"] = parent_id
        if label is not None:
            params["label"] = label
        if ids is not None:
            params["ids"] = ",".join(str(i) for i in ids)
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Task.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def filter_tasks(
        self,
        *,
        query: Annotated[str, MaxLen(1024)] | None = None,
        lang: str | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Task]]:
        """
        Get an iterable of lists of active tasks matching the filter.

        The response is an iterable of lists of active tasks matching the criteria.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param query: Query tasks using Todoist's filter language.
        :param lang: Language for task content (e.g., 'en').
        :param limit: Maximum number of tasks per page.
        :return: An iterable of lists of tasks.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(TASKS_FILTER_PATH)

        params: dict[str, Any] = {}
        if query is not None:
            params["query"] = query
        if lang is not None:
            params["lang"] = lang
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Task.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def add_task(  # noqa: PLR0912
        self,
        content: Annotated[str, MinLen(1), MaxLen(500)],
        *,
        description: Annotated[str, MaxLen(16383)] | None = None,
        project_id: str | None = None,
        section_id: str | None = None,
        parent_id: str | None = None,
        labels: list[Annotated[str, MaxLen(100)]] | None = None,
        priority: Annotated[int, Ge(1), Le(4)] | None = None,
        due_string: Annotated[str, MaxLen(150)] | None = None,
        due_lang: LanguageCode | None = None,
        due_date: date | None = None,
        due_datetime: datetime | None = None,
        assignee_id: str | None = None,
        order: int | None = None,
        auto_reminder: bool | None = None,
        auto_parse_labels: bool | None = None,
        duration: Annotated[int, Ge(1)] | None = None,
        duration_unit: Literal["minute", "day"] | None = None,
        deadline_date: date | None = None,
        deadline_lang: LanguageCode | None = None,
    ) -> Task:
        """
        Create a new task.

        :param content: The text content of the task.
        :param project_id: The ID of the project to add the task to.
        :param section_id: The ID of the section to add the task to.
        :param parent_id: The ID of the parent task.
        :param labels: The task's labels (a list of names).
        :param priority: The priority of the task (4 for very urgent).
        :param due_string: The due date in natural language format.
        :param due_lang: Language for parsing the due date (e.g., 'en').
        :param due_date: The due date as a date object.
        :param due_datetime: The due date and time as a datetime object.
        :param assignee_id: User ID to whom the task is assigned.
        :param description: Description for the task.
        :param order: The order of task in the project or section.
        :param auto_reminder: Whether to add default reminder if date with time is set.
        :param auto_parse_labels: Whether to parse labels from task content.
        :param duration: The amount of time the task will take.
        :param duration_unit: The unit of time for duration.
        :param deadline_date: The deadline date as a date object.
        :param deadline_lang: Language for parsing the deadline date.
        :return: The newly created task.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Task dictionary.
        """
        endpoint = get_api_url(TASKS_PATH)

        data: dict[str, Any] = {"content": content}
        if description is not None:
            data["description"] = description
        if project_id is not None:
            data["project_id"] = project_id
        if section_id is not None:
            data["section_id"] = section_id
        if parent_id is not None:
            data["parent_id"] = parent_id
        if labels is not None:
            data["labels"] = labels
        if priority is not None:
            data["priority"] = priority
        if due_string is not None:
            data["due_string"] = due_string
        if due_lang is not None:
            data["due_lang"] = due_lang
        if due_date is not None:
            data["due_date"] = format_date(due_date)
        if due_datetime is not None:
            data["due_datetime"] = format_datetime(due_datetime)
        if assignee_id is not None:
            data["assignee_id"] = assignee_id
        if order is not None:
            data["order"] = order
        if auto_reminder is not None:
            data["auto_reminder"] = auto_reminder
        if auto_parse_labels is not None:
            data["auto_parse_labels"] = auto_parse_labels
        if duration is not None:
            data["duration"] = duration
        if duration_unit is not None:
            data["duration_unit"] = duration_unit
        if deadline_date is not None:
            data["deadline_date"] = format_date(deadline_date)
        if deadline_lang is not None:
            data["deadline_lang"] = deadline_lang

        task_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Task.from_dict(task_data)

    def add_task_quick(
        self,
        text: str,
        *,
        note: str | None = None,
        reminder: str | None = None,
        auto_reminder: bool = True,
    ) -> Task:
        """
        Create a new task using Todoist's Quick Add syntax.

        This automatically parses dates, deadlines, projects, labels, priorities, etc,
        from the provided text (e.g., "Buy milk #Shopping @groceries tomorrow p1").

        :param text: The task text using Quick Add syntax.
        :param note: Optional note to be added to the task.
        :param reminder: Optional reminder date in free form text.
        :param auto_reminder: Whether to add default reminder if date with time is set.
        :return: A result object containing the parsed task data and metadata.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response cannot be parsed into a QuickAddResult.
        """
        endpoint = get_api_url(TASKS_QUICK_ADD_PATH)

        data = {
            "meta": True,
            "text": text,
            "auto_reminder": auto_reminder,
        }

        if note is not None:
            data["note"] = note
        if reminder is not None:
            data["reminder"] = reminder

        task_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Task.from_dict(task_data)

    def update_task(  # noqa: PLR0912
        self,
        task_id: str,
        *,
        content: Annotated[str, MinLen(1), MaxLen(500)] | None = None,
        description: Annotated[str, MaxLen(16383)] | None = None,
        labels: list[Annotated[str, MaxLen(60)]] | None = None,
        priority: Annotated[int, Ge(1), Le(4)] | None = None,
        due_string: Annotated[str, MaxLen(150)] | None = None,
        due_lang: LanguageCode | None = None,
        due_date: date | None = None,
        due_datetime: datetime | None = None,
        assignee_id: str | None = None,
        day_order: int | None = None,
        collapsed: bool | None = None,
        duration: Annotated[int, Ge(1)] | None = None,
        duration_unit: Literal["minute", "day"] | None = None,
        deadline_date: date | None = None,
        deadline_lang: LanguageCode | None = None,
    ) -> Task:
        """
        Update an existing task.

        Only the fields to be updated need to be provided.

        :param task_id: The ID of the task to update.
        :param content: The text content of the task.
        :param description: Description for the task.
        :param labels: The task's labels (a list of names).
        :param priority: The priority of the task (4 for very urgent).
        :param due_string: The due date in natural language format.
        :param due_lang: Language for parsing the due date (e.g., 'en').
        :param due_date: The due date as a date object.
        :param due_datetime: The due date and time as a datetime object.
        :param assignee_id: User ID to whom the task is assigned.
        :param day_order: The order of the task inside Today or Next 7 days view.
        :param collapsed: Whether the task's sub-tasks are collapsed.
        :param duration: The amount of time the task will take.
        :param duration_unit: The unit of time for duration.
        :param deadline_date: The deadline date as a date object.
        :param deadline_lang: Language for parsing the deadline date.
        :return: the updated Task.
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}")

        data: dict[str, Any] = {}
        if content is not None:
            data["content"] = content
        if description is not None:
            data["description"] = description
        if labels is not None:
            data["labels"] = labels
        if priority is not None:
            data["priority"] = priority
        if due_string is not None:
            data["due_string"] = due_string
        if due_lang is not None:
            data["due_lang"] = due_lang
        if due_date is not None:
            data["due_date"] = format_date(due_date)
        if due_datetime is not None:
            data["due_datetime"] = format_datetime(due_datetime)
        if assignee_id is not None:
            data["assignee_id"] = assignee_id
        if day_order is not None:
            data["day_order"] = day_order
        if collapsed is not None:
            data["collapsed"] = collapsed
        if duration is not None:
            data["duration"] = duration
        if duration_unit is not None:
            data["duration_unit"] = duration_unit
        if deadline_date is not None:
            data["deadline_date"] = format_date(deadline_date)
        if deadline_lang is not None:
            data["deadline_lang"] = deadline_lang

        task_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Task.from_dict(task_data)

    def complete_task(self, task_id: str) -> bool:
        """
        Complete a task.

        For recurring tasks, this schedules the next occurrence.
        For non-recurring tasks, it marks them as completed.

        :param task_id: The ID of the task to close.
        :return: True if the task was closed successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}/close")
        return post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def uncomplete_task(self, task_id: str) -> bool:
        """
        Uncomplete a (completed) task.

        Any parent tasks or sections will also be uncompleted.

        :param task_id: The ID of the task to reopen.
        :return: True if the task was uncompleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}/reopen")
        return post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def move_task(
        self,
        task_id: str,
        project_id: str | None = None,
        section_id: str | None = None,
        parent_id: str | None = None,
    ) -> bool:
        """
        Move a task to a different project, section, or parent task.

        `project_id` takes predence, followed by
        `section_id` (which also updates `project_id`),
        and then `parent_id` (which also updates `section_id` and `project_id`).

        :param task_id: The ID of the task to move.
        :param project_id: The ID of the project to move the task to.
        :param section_id: The ID of the section to move the task to.
        :param parent_id: The ID of the parent to move the task to.
        :return: True if the task was moved successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises ValueError: If neither `project_id`, `section_id`,
                nor `parent_id` is provided.
        """
        if project_id is None and section_id is None and parent_id is None:
            raise ValueError(
                "Either `project_id`, `section_id`, or `parent_id` must be provided."
            )

        data: dict[str, Any] = {}
        if project_id is not None:
            data["project_id"] = project_id
        if section_id is not None:
            data["section_id"] = section_id
        if parent_id is not None:
            data["parent_id"] = parent_id
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}/move")
        return post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )

    def delete_task(self, task_id: str) -> bool:
        """
        Delete a task.

        :param task_id: The ID of the task to delete.
        :return: True if the task was deleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{TASKS_PATH}/{task_id}")
        return delete(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def get_completed_tasks_by_due_date(
        self,
        *,
        since: datetime,
        until: datetime,
        workspace_id: str | None = None,
        project_id: str | None = None,
        section_id: str | None = None,
        parent_id: str | None = None,
        filter_query: str | None = None,
        filter_lang: str | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Task]]:
        """
        Get an iterable of lists of completed tasks within a due date range.

        Retrieves tasks completed within a specific due date range (up to 6 weeks).
        Supports filtering by workspace, project, section, parent task, or a query.

        The response is an iterable of lists of completed tasks. Be aware that each
        iteration fires off a network request to the Todoist API, and may result in
        rate limiting or other API restrictions.

        :param since: Start of the date range (inclusive).
        :param until: End of the date range (inclusive).
        :param workspace_id: Filter by workspace ID.
        :param project_id: Filter by project ID.
        :param section_id: Filter by section ID.
        :param parent_id: Filter by parent task ID.
        :param filter_query: Filter by a query string.
        :param filter_lang: Language for the filter query (e.g., 'en').
        :param limit: Maximum number of tasks per page (default 50).
        :return: An iterable of lists of completed tasks.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(TASKS_COMPLETED_BY_DUE_DATE_PATH)

        params: dict[str, Any] = {
            "since": format_datetime(since),
            "until": format_datetime(until),
        }
        if workspace_id is not None:
            params["workspace_id"] = workspace_id
        if project_id is not None:
            params["project_id"] = project_id
        if section_id is not None:
            params["section_id"] = section_id
        if parent_id is not None:
            params["parent_id"] = parent_id
        if filter_query is not None:
            params["filter_query"] = filter_query
        if filter_lang is not None:
            params["filter_lang"] = filter_lang
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "items",
            Task.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def get_completed_tasks_by_completion_date(
        self,
        *,
        since: datetime,
        until: datetime,
        workspace_id: str | None = None,
        filter_query: str | None = None,
        filter_lang: str | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Task]]:
        """
        Get an iterable of lists of completed tasks within a date range.

        Retrieves tasks completed within a specific date range (up to 3 months).
        Supports filtering by workspace or a filter query.

        The response is an iterable of lists of completed tasks. Be aware that each
        iteration fires off a network request to the Todoist API, and may result in
        rate limiting or other API restrictions.

        :param since: Start of the date range (inclusive).
        :param until: End of the date range (inclusive).
        :param workspace_id: Filter by workspace ID.
        :param filter_query: Filter by a query string.
        :param filter_lang: Language for the filter query (e.g., 'en').
        :param limit: Maximum number of tasks per page (default 50).
        :return: An iterable of lists of completed tasks.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(TASKS_COMPLETED_BY_COMPLETION_DATE_PATH)

        params: dict[str, Any] = {
            "since": format_datetime(since),
            "until": format_datetime(until),
        }
        if workspace_id is not None:
            params["workspace_id"] = workspace_id
        if filter_query is not None:
            params["filter_query"] = filter_query
        if filter_lang is not None:
            params["filter_lang"] = filter_lang
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "items",
            Task.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def get_project(self, project_id: str) -> Project:
        """
        Get a project by its ID.

        :param project_id: The ID of the project to retrieve.
        :return: The requested project.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Project dictionary.
        """
        endpoint = get_api_url(f"{PROJECTS_PATH}/{project_id}")
        project_data: dict[str, Any] = get(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Project.from_dict(project_data)

    def get_projects(
        self,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Project]]:
        """
        Get an iterable of lists of active projects.

        The response is an iterable of lists of active projects.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param limit: Maximum number of projects per page.
        :return: An iterable of lists of projects.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(PROJECTS_PATH)
        params: dict[str, Any] = {}
        if limit is not None:
            params["limit"] = limit
        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Project.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def search_projects(
        self,
        query: Annotated[str, MinLen(1), MaxLen(1024)],
        *,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Project]]:
        """
        Search active projects by name.

        The response is an iterable of lists of projects matching the query.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param query: Query string for project names.
        :param limit: Maximum number of projects per page.
        :return: An iterable of lists of projects.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(f"{PROJECTS_PATH}/{PROJECTS_SEARCH_PATH_SUFFIX}")

        params: dict[str, Any] = {"query": query}
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Project.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def add_project(
        self,
        name: Annotated[str, MinLen(1), MaxLen(120)],
        *,
        description: Annotated[str, MaxLen(16383)] | None = None,
        parent_id: str | None = None,
        color: ColorString | None = None,
        is_favorite: bool | None = None,
        view_style: ViewStyle | None = None,
    ) -> Project:
        """
        Create a new project.

        :param name: The name of the project.
        :param description: Description for the project (up to 1024 characters).
        :param parent_id: The ID of the parent project. Set to null for root projects.
        :param color: The color of the project icon.
        :param is_favorite: Whether the project is a favorite.
        :param view_style: A string value (either 'list' or 'board', default is 'list').
        :return: The newly created project.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Project dictionary.
        """
        endpoint = get_api_url(PROJECTS_PATH)

        data: dict[str, Any] = {"name": name}
        if parent_id is not None:
            data["parent_id"] = parent_id
        if description is not None:
            data["description"] = description
        if color is not None:
            data["color"] = color
        if is_favorite is not None:
            data["is_favorite"] = is_favorite
        if view_style is not None:
            data["view_style"] = view_style

        project_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Project.from_dict(project_data)

    def update_project(
        self,
        project_id: str,
        *,
        name: Annotated[str, MinLen(1), MaxLen(120)] | None = None,
        description: Annotated[str, MaxLen(16383)] | None = None,
        color: ColorString | None = None,
        is_favorite: bool | None = None,
        view_style: ViewStyle | None = None,
    ) -> Project:
        """
        Update an existing project.

        Only the fields to be updated need to be provided as keyword arguments.

        :param project_id: The ID of the project to update.
        :param name: The name of the project.
        :param description: Description for the project (up to 1024 characters).
        :param color: The color of the project icon.
        :param is_favorite: Whether the project is a favorite.
        :param view_style: A string value (either 'list' or 'board').
        :return: the updated Project.
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{PROJECTS_PATH}/{project_id}")

        data: dict[str, Any] = {}

        if name is not None:
            data["name"] = name
        if description is not None:
            data["description"] = description
        if color is not None:
            data["color"] = color
        if is_favorite is not None:
            data["is_favorite"] = is_favorite
        if view_style is not None:
            data["view_style"] = view_style

        project_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Project.from_dict(project_data)

    def archive_project(self, project_id: str) -> Project:
        """
        Archive a project.

        For personal projects, archives it only for the user.
        For workspace projects, archives it for all members.

        :param project_id: The ID of the project to archive.
        :return: The archived project object.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Project dictionary.
        """
        endpoint = get_api_url(
            f"{PROJECTS_PATH}/{project_id}/{PROJECT_ARCHIVE_PATH_SUFFIX}"
        )
        project_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Project.from_dict(project_data)

    def unarchive_project(self, project_id: str) -> Project:
        """
        Unarchive a project.

        Restores a previously archived project.

        :param project_id: The ID of the project to unarchive.
        :return: The unarchived project object.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Project dictionary.
        """
        endpoint = get_api_url(
            f"{PROJECTS_PATH}/{project_id}/{PROJECT_UNARCHIVE_PATH_SUFFIX}"
        )
        project_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Project.from_dict(project_data)

    def delete_project(self, project_id: str) -> bool:
        """
        Delete a project.

        All nested sections and tasks will also be deleted.

        :param project_id: The ID of the project to delete.
        :return: True if the project was deleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{PROJECTS_PATH}/{project_id}")
        return delete(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def get_collaborators(
        self,
        project_id: str,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Collaborator]]:
        """
        Get an iterable of lists of collaborators in shared projects.

        The response is an iterable of lists of collaborators in shared projects,
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param project_id: The ID of the project.
        :param limit: Maximum number of collaborators per page.
        :return: An iterable of lists of collaborators.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(f"{PROJECTS_PATH}/{project_id}/{COLLABORATORS_PATH}")
        params: dict[str, Any] = {}
        if limit is not None:
            params["limit"] = limit
        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Collaborator.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def get_section(self, section_id: str) -> Section:
        """
        Get a specific section by its ID.

        :param section_id: The ID of the section to retrieve.
        :return: The requested section.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Section dictionary.
        """
        endpoint = get_api_url(f"{SECTIONS_PATH}/{section_id}")
        section_data: dict[str, Any] = get(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Section.from_dict(section_data)

    def get_sections(
        self,
        project_id: str | None = None,
        *,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Section]]:
        """
        Get an iterable of lists of active sections.

        Supports filtering by `project_id` and pagination arguments.

        The response is an iterable of lists of active sections.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param project_id: Filter sections by project ID.
        :param limit: Maximum number of sections per page.
        :return: An iterable of lists of sections.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(SECTIONS_PATH)

        params: dict[str, Any] = {}
        if project_id is not None:
            params["project_id"] = project_id
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Section.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def search_sections(
        self,
        query: Annotated[str, MinLen(1), MaxLen(1024)],
        *,
        project_id: str | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Section]]:
        """
        Search active sections by name.

        The response is an iterable of lists of sections matching the query.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param query: Query string for section names.
        :param project_id: If set, search sections within the given project only.
        :param limit: Maximum number of sections per page.
        :return: An iterable of lists of sections.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(f"{SECTIONS_PATH}/{SECTIONS_SEARCH_PATH_SUFFIX}")

        params: dict[str, Any] = {"query": query}
        if project_id is not None:
            params["project_id"] = project_id
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Section.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def add_section(
        self,
        name: Annotated[str, MinLen(1), MaxLen(2048)],
        project_id: str,
        *,
        order: int | None = None,
    ) -> Section:
        """
        Create a new section within a project.

        :param name: The name of the section.
        :param project_id: The ID of the project to add the section to.
        :param order: The order of the section among all sections in the project.
        :return: The newly created section.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Section dictionary.
        """
        endpoint = get_api_url(SECTIONS_PATH)

        data: dict[str, Any] = {"name": name, "project_id": project_id}
        if order is not None:
            data["order"] = order

        section_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Section.from_dict(section_data)

    def update_section(
        self,
        section_id: str,
        name: Annotated[str, MinLen(1), MaxLen(2048)],
    ) -> Section:
        """
        Update an existing section.

        Currently, only `name` can be updated.

        :param section_id: The ID of the section to update.
        :param name: The new name for the section.
        :return: the updated Section.
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{SECTIONS_PATH}/{section_id}")
        section_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data={"name": name},
        )
        return Section.from_dict(section_data)

    def delete_section(self, section_id: str) -> bool:
        """
        Delete a section.

        All tasks within the section will also be deleted.

        :param section_id: The ID of the section to delete.
        :return: True if the section was deleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{SECTIONS_PATH}/{section_id}")
        return delete(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def get_comment(self, comment_id: str) -> Comment:
        """
        Get a specific comment by its ID.

        :param comment_id: The ID of the comment to retrieve.
        :return: The requested comment.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Comment dictionary.
        """
        endpoint = get_api_url(f"{COMMENTS_PATH}/{comment_id}")
        comment_data: dict[str, Any] = get(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Comment.from_dict(comment_data)

    def get_comments(
        self,
        *,
        project_id: str | None = None,
        task_id: str | None = None,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Comment]]:
        """
        Get an iterable of lists of comments for a task or project.

        Requires either `project_id` or `task_id` to be set.

        The response is an iterable of lists of comments.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param project_id: The ID of the project to retrieve comments for.
        :param task_id: The ID of the task to retrieve comments for.
        :param limit: Maximum number of comments per page.
        :return: An iterable of lists of comments.
        :raises ValueError: If neither `project_id` nor `task_id` is provided.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        if project_id is None and task_id is None:
            raise ValueError("Either `project_id` or `task_id` must be provided.")

        endpoint = get_api_url(COMMENTS_PATH)

        params: dict[str, Any] = {}
        if project_id is not None:
            params["project_id"] = project_id
        if task_id is not None:
            params["task_id"] = task_id
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Comment.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def add_comment(
        self,
        content: Annotated[str, MaxLen(15000)],
        *,
        project_id: str | None = None,
        task_id: str | None = None,
        attachment: Attachment | None = None,
        uids_to_notify: list[str] | None = None,
    ) -> Comment:
        """
        Create a new comment on a task or project.

        Requires either `project_id` or `task_id` to be set,
        and can optionally include an `attachment` object.

        :param content: The text content of the comment (supports Markdown).
        :param project_id: The ID of the project to add the comment to.
        :param task_id: The ID of the task to add the comment to.
        :param attachment: The attachment object to include with the comment.
        :param uids_to_notify: A list of user IDs to notify.
        :return: The newly created comment.
        :raises ValueError: If neither `project_id` nor `task_id` is provided.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Comment dictionary.
        """
        if project_id is None and task_id is None:
            raise ValueError("Either `project_id` or `task_id` must be provided.")

        endpoint = get_api_url(COMMENTS_PATH)

        data: dict[str, Any] = {"content": content}
        if project_id is not None:
            data["project_id"] = project_id
        if task_id is not None:
            data["task_id"] = task_id
        if attachment is not None:
            data["attachment"] = attachment.to_dict()
        if uids_to_notify is not None:
            data["uids_to_notify"] = uids_to_notify

        comment_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Comment.from_dict(comment_data)

    def update_comment(
        self, comment_id: str, content: Annotated[str, MaxLen(15000)]
    ) -> Comment:
        """
        Update an existing comment.

        Currently, only `content` can be updated.

        :param comment_id: The ID of the comment to update.
        :param content: The new text content for the comment.
        :return: the updated Comment.
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{COMMENTS_PATH}/{comment_id}")
        comment_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data={"content": content},
        )
        return Comment.from_dict(comment_data)

    def delete_comment(self, comment_id: str) -> bool:
        """
        Delete a comment.

        :param comment_id: The ID of the comment to delete.
        :return: True if the comment was deleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{COMMENTS_PATH}/{comment_id}")
        return delete(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def get_label(self, label_id: str) -> Label:
        """
        Get a specific personal label by its ID.

        :param label_id: The ID of the label to retrieve.
        :return: The requested label.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Label dictionary.
        """
        endpoint = get_api_url(f"{LABELS_PATH}/{label_id}")
        label_data: dict[str, Any] = get(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )
        return Label.from_dict(label_data)

    def get_labels(
        self,
        *,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Label]]:
        """
        Get an iterable of lists of personal labels.

        Supports pagination arguments.

        The response is an iterable of lists of personal labels.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param limit: Maximum number of labels per page.
        :return: An iterable of lists of personal labels.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(LABELS_PATH)

        params: dict[str, Any] = {}
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Label.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def search_labels(
        self,
        query: Annotated[str, MinLen(1), MaxLen(1024)],
        *,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[Label]]:
        """
        Search personal labels by name.

        The response is an iterable of lists of labels matching the query.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param query: Query string for label names.
        :param limit: Maximum number of labels per page.
        :return: An iterable of lists of labels.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(f"{LABELS_PATH}/{LABELS_SEARCH_PATH_SUFFIX}")

        params: dict[str, Any] = {"query": query}
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            Label.from_dict,
            self._token,
            self._request_id_fn,
            params,
        )

    def add_label(
        self,
        name: Annotated[str, MinLen(1), MaxLen(60)],
        *,
        color: ColorString | None = None,
        item_order: int | None = None,
        is_favorite: bool | None = None,
    ) -> Label:
        """
        Create a new personal label.

        :param name: The name of the label.
        :param color: The color of the label icon.
        :param item_order: Label's order in the label list.
        :param is_favorite: Whether the label is a favorite.
        :return: The newly created label.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response is not a valid Label dictionary.
        """
        endpoint = get_api_url(LABELS_PATH)

        data: dict[str, Any] = {"name": name}

        if color is not None:
            data["color"] = color
        if item_order is not None:
            data["item_order"] = item_order
        if is_favorite is not None:
            data["is_favorite"] = is_favorite

        label_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Label.from_dict(label_data)

    def update_label(
        self,
        label_id: str,
        *,
        name: Annotated[str, MinLen(1), MaxLen(60)] | None = None,
        color: ColorString | None = None,
        item_order: int | None = None,
        is_favorite: bool | None = None,
    ) -> Label:
        """
        Update a personal label.

        Only the fields to be updated need to be provided as keyword arguments.

        :param label_id: The ID of the label.
        :param name: The name of the label.
        :param color: The color of the label icon.
        :param item_order: Label's order in the label list.
        :param is_favorite: Whether the label is a favorite.
        :return: the updated Label.
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{LABELS_PATH}/{label_id}")

        data: dict[str, Any] = {}
        if name is not None:
            data["name"] = name
        if color is not None:
            data["color"] = color
        if item_order is not None:
            data["item_order"] = item_order
        if is_favorite is not None:
            data["is_favorite"] = is_favorite

        label_data: dict[str, Any] = post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )
        return Label.from_dict(label_data)

    def delete_label(self, label_id: str) -> bool:
        """
        Delete a personal label.

        Instances of the label will be removed from tasks.

        :param label_id: The ID of the label to delete.
        :return: True if the label was deleted successfully,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(f"{LABELS_PATH}/{label_id}")
        return delete(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
        )

    def get_shared_labels(
        self,
        *,
        omit_personal: bool = False,
        limit: Annotated[int, Ge(1), Le(200)] | None = None,
    ) -> Iterator[list[str]]:
        """
        Get an iterable of lists of shared label names.

        Includes labels from collaborators on shared projects that are not in the
        user's personal labels. Can optionally exclude personal label names using
        `omit_personal=True`. Supports pagination arguments.

        The response is an iterable of lists of shared label names.
        Be aware that each iteration fires off a network request to the Todoist API,
        and may result in rate limiting or other API restrictions.

        :param omit_personal: Optional boolean flag to omit personal label names.
        :param limit: Maximum number of labels per page.
        :return: An iterable of lists of shared label names (strings).
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        endpoint = get_api_url(SHARED_LABELS_PATH)

        params: dict[str, Any] = {"omit_personal": omit_personal}
        if limit is not None:
            params["limit"] = limit

        return ResultsPaginator(
            self._session,
            endpoint,
            "results",
            str,
            self._token,
            self._request_id_fn,
            params,
        )

    def rename_shared_label(
        self,
        name: Annotated[str, MaxLen(60)],
        new_name: Annotated[str, MinLen(1), MaxLen(60)],
    ) -> bool:
        """
        Rename all occurrences of a shared label across all projects.

        :param name: The current name of the shared label to rename.
        :param new_name: The new name for the shared label.
        :return: True if the rename was successful,
                 False otherwise (possibly raise `HTTPError` instead).
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(SHARED_LABELS_RENAME_PATH)
        return post(
            self._session,
            endpoint,
            self._token,
            params={"name": name},
            data={"new_name": new_name},
        )

    def remove_shared_label(self, name: Annotated[str, MaxLen(60)]) -> bool:
        """
        Remove all occurrences of a shared label across all projects.

        This action removes the label string from all tasks where it appears.

        :param name: The name of the shared label to remove.
        :return: True if the removal was successful,
        :raises requests.exceptions.HTTPError: If the API request fails.
        """
        endpoint = get_api_url(SHARED_LABELS_REMOVE_PATH)
        data = {"name": name}
        return post(
            self._session,
            endpoint,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            data=data,
        )


T = TypeVar("T")


class ResultsPaginator(Iterator[list[T]]):
    """
    Iterator for paginated results from the Todoist API.

    It encapsulates the logic for fetching and iterating through paginated results
    from Todoist API endpoints. It handles cursor-based pagination automatically,
    requesting new pages as needed when iterating.
    """

    _session: requests.Session
    _url: str
    _results_field: str
    _results_inst: Callable[[Any], T]
    _token: str
    _cursor: str | None

    def __init__(
        self,
        session: requests.Session,
        url: str,
        results_field: str,
        results_inst: Callable[[Any], T],
        token: str,
        request_id_fn: Callable[[], str] | None,
        params: dict[str, Any],
    ) -> None:
        """
        Initialize the ResultsPaginator.

        :param session: The requests Session to use for API calls.
        :param url: The API endpoint URL to fetch results from.
        :param results_field: The key in the API response that contains the results.
        :param results_inst: A callable that converts result items to objects of type T.
        :param token: The authentication token for the Todoist API.
        :param params: Query parameters to include in API requests.
        """
        self._session = session
        self._url = url
        self._results_field = results_field
        self._results_inst = results_inst
        self._token = token
        self._request_id_fn = request_id_fn
        self._params = params
        self._cursor = ""  # empty string for first page

    def __next__(self) -> list[T]:
        """
        Fetch and return the next page of results from the Todoist API.

        :return: A list of results.
        :raises requests.exceptions.HTTPError: If the API request fails.
        :raises TypeError: If the API response structure is unexpected.
        """
        if self._cursor is None:
            raise StopIteration

        params = self._params.copy()
        if self._cursor != "":
            params["cursor"] = self._cursor

        data: dict[str, Any] = get(
            self._session,
            self._url,
            self._token,
            self._request_id_fn() if self._request_id_fn else None,
            params,
        )
        self._cursor = data.get("next_cursor")

        results: list[Any] = data.get(self._results_field, [])
        return [self._results_inst(result) for result in results]



from __future__ import annotations

from typing import Any, TypedDict


class PaginatedResults(TypedDict):
    results: list[dict[str, Any]]
    next_cursor: str | None


class PaginatedItems(TypedDict):
    items: list[dict[str, Any]]
    next_cursor: str | None


DEFAULT_API_URL = "https://api.todoist.com/api/v1"
DEFAULT_OAUTH_URL = "https://todoist.com/oauth"

DEFAULT_TOKEN = "some-default-token"

DEFAULT_REQUEST_ID = "f00dbeef-cafe-4bad-a555-deadc0decafe"

DEFAULT_DUE_RESPONSE = {
    "date": "2016-09-01",
    "timezone": "Europe/Moscow",
    "string": "tomorrow at 12",
    "lang": "en",
    "is_recurring": True,
}

DEFAULT_DEADLINE_RESPONSE = {
    "date": "2016-09-01",
    "lang": "en",
}

DEFAULT_DURATION_RESPONSE = {
    "amount": 60,
    "unit": "minute",
}

DEFAULT_META_RESPONSE: dict[str, Any] = {
    "project": ["6X7rM8997g3RQmvh", "Inbox"],
    "section": [None, None],
    "assignee": [None, None],
    "labels": {},
    "due": None,
    "deadline": None,
}

DEFAULT_PROJECT_RESPONSE = {
    "id": "6X7rM8997g3RQmvh",
    "name": "Inbox",
    "description": "",
    "parent_id": "6X7rfFVPjhvv84XG",
    "folder_id": None,
    "workspace_id": None,
    "child_order": 1,
    "color": "red",
    "shared": False,
    "collapsed": False,
    "is_favorite": False,
    "is_inbox_project": True,
    "can_assign_tasks": False,
    "is_archived": False,
    "view_style": "list",
    "created_at": "2023-02-01T00:00:00.000000Z",
    "updated_at": "2025-04-03T03:14:15.926536Z",
}

DEFAULT_PROJECT_RESPONSE_2 = dict(DEFAULT_PROJECT_RESPONSE)
DEFAULT_PROJECT_RESPONSE_2["id"] = "6X7rfFVPjhvv84XG"
DEFAULT_PROJECT_RESPONSE_2["is_inbox_project"] = False


DEFAULT_PROJECT_RESPONSE_3 = dict(DEFAULT_PROJECT_RESPONSE)
DEFAULT_PROJECT_RESPONSE_3["id"] = "6X7rfEVP8hvv25ZQ"
DEFAULT_PROJECT_RESPONSE_3["is_inbox_project"] = False

DEFAULT_PROJECTS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_PROJECT_RESPONSE, DEFAULT_PROJECT_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_PROJECT_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_TASK_RESPONSE: dict[str, Any] = {
    "id": "6X7rM8997g3RQmvh",
    "content": "Some task content",
    "description": "Some task description",
    "project_id": "6Jf8VQXxpwv56VQ7",
    "section_id": "3Ty8VQXxpwv28PK3",
    "parent_id": "6X7rf9x6pv2FGghW",
    "labels": [],
    "priority": 1,
    "due": DEFAULT_DUE_RESPONSE,
    "deadline": DEFAULT_DEADLINE_RESPONSE,
    "duration": DEFAULT_DURATION_RESPONSE,
    "collapsed": False,
    "child_order": 3,
    "responsible_uid": "2423523",
    "assigned_by_uid": "2971358",
    "completed_at": None,
    "added_by_uid": "34567",
    "added_at": "2014-09-26T08:25:05.000000Z",
    "updated_at": "2016-01-02T21:00:30.000000Z",
}

DEFAULT_TASK_RESPONSE_2 = dict(DEFAULT_TASK_RESPONSE)
DEFAULT_TASK_RESPONSE_2["id"] = "6X7rfFVPjhvv84XG"

DEFAULT_TASK_RESPONSE_3 = dict(DEFAULT_TASK_RESPONSE)
DEFAULT_TASK_RESPONSE_3["id"] = "6X7rF9xvX25jTxm5"

DEFAULT_TASKS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_TASK_RESPONSE, DEFAULT_TASK_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_TASK_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_TASK_META_RESPONSE = dict(DEFAULT_TASK_RESPONSE)
DEFAULT_TASK_META_RESPONSE["meta"] = DEFAULT_META_RESPONSE

DEFAULT_COMPLETED_TASK_RESPONSE = dict(DEFAULT_TASK_RESPONSE)
DEFAULT_COMPLETED_TASK_RESPONSE["completed_at"] = "2024-02-13T10:00:00.000000Z"

DEFAULT_COMPLETED_TASK_RESPONSE_2 = dict(DEFAULT_COMPLETED_TASK_RESPONSE)
DEFAULT_COMPLETED_TASK_RESPONSE_2["id"] = "6X7rfFVPjhvv84XG"

DEFAULT_COMPLETED_TASK_RESPONSE_3 = dict(DEFAULT_COMPLETED_TASK_RESPONSE)
DEFAULT_COMPLETED_TASK_RESPONSE_3["id"] = "6X7rfEVP8hvv25ZQ"

DEFAULT_COMPLETED_TASKS_RESPONSE: list[PaginatedItems] = [
    {
        "items": [
            DEFAULT_COMPLETED_TASK_RESPONSE,
            DEFAULT_COMPLETED_TASK_RESPONSE_2,
        ],
        "next_cursor": "next",
    },
    {
        "items": [DEFAULT_COMPLETED_TASK_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_COLLABORATOR_RESPONSE = {
    "id": "6X7rM8997g3RQmvh",
    "name": "Alice",
    "email": "alice@example.com",
}

DEFAULT_COLLABORATOR_RESPONSE_2 = dict(DEFAULT_COLLABORATOR_RESPONSE)
DEFAULT_COLLABORATOR_RESPONSE_2["id"] = "6X7rfFVPjhvv84XG"

DEFAULT_COLLABORATOR_RESPONSE_3 = dict(DEFAULT_COLLABORATOR_RESPONSE)
DEFAULT_COLLABORATOR_RESPONSE_3["id"] = "6X7rjKtP98vG84rK"

DEFAULT_COLLABORATORS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_COLLABORATOR_RESPONSE, DEFAULT_COLLABORATOR_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_COLLABORATOR_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_SECTION_RESPONSE = {
    "id": "6X7rM8997g3RQmvh",
    "project_id": "4567",
    "name": "A Section",
    "collapsed": False,
    "order": 1,
}

DEFAULT_SECTION_RESPONSE_2 = dict(DEFAULT_SECTION_RESPONSE)
DEFAULT_SECTION_RESPONSE_2["id"] = "6X7FxXvX84jHphx"

DEFAULT_SECTION_RESPONSE_3 = dict(DEFAULT_SECTION_RESPONSE)
DEFAULT_SECTION_RESPONSE_3["id"] = "6X7rF9xvX25jTzm7"

DEFAULT_SECTIONS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_SECTION_RESPONSE, DEFAULT_SECTION_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_SECTION_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_ATTACHMENT_RESPONSE = {
    "resource_type": "file",
    "file_name": "File.pdf",
    "file_type": "application/pdf",
    "file_size": 4321,
    "file_url": "https://cdn-domain.tld/path/to/file.pdf",
    "upload_state": "completed",
    "image": "https://cdn-domain.tld/path/to/some_image.jpg",
    "image_width": 800,
    "image_height": 600,
    "url": "https://todoist.com",
    "title": "Todoist Website",
}

DEFAULT_COMMENT_RESPONSE: dict[str, Any] = {
    "id": "6X7rM8997g3RQmvh",
    "content": "A comment",
    "posted_uid": "34567",
    "posted_at": "2019-09-22T07:00:00.000000Z",
    "task_id": "6X7rM8997g3RQmvh",
    "project_id": "6X7rfEVP8hvv25ZQ",
    "attachment": DEFAULT_ATTACHMENT_RESPONSE,
}

DEFAULT_COMMENT_RESPONSE_2 = dict(DEFAULT_COMMENT_RESPONSE)
DEFAULT_COMMENT_RESPONSE_2["id"] = "6X7rfFVPjhvv84XG"
DEFAULT_COMMENT_RESPONSE_2["attachment"] = None

DEFAULT_COMMENT_RESPONSE_3 = dict(DEFAULT_COMMENT_RESPONSE)
DEFAULT_COMMENT_RESPONSE_3["id"] = "6X7rfFVPjhvv65HG"
DEFAULT_COMMENT_RESPONSE_3["attachment"] = None

DEFAULT_COMMENTS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_COMMENT_RESPONSE, DEFAULT_COMMENT_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_COMMENT_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_LABEL_RESPONSE = {
    "id": "1234",
    "name": "A label",
    "color": "red",
    "order": 1,
    "is_favorite": True,
}

DEFAULT_LABEL_RESPONSE_2 = dict(DEFAULT_LABEL_RESPONSE)
DEFAULT_LABEL_RESPONSE_2["id"] = "4567"

DEFAULT_LABEL_RESPONSE_3 = dict(DEFAULT_LABEL_RESPONSE)
DEFAULT_LABEL_RESPONSE_3["id"] = "6789"

DEFAULT_LABELS_RESPONSE: list[PaginatedResults] = [
    {
        "results": [DEFAULT_LABEL_RESPONSE, DEFAULT_LABEL_RESPONSE_2],
        "next_cursor": "next",
    },
    {
        "results": [DEFAULT_LABEL_RESPONSE_3],
        "next_cursor": None,
    },
]

DEFAULT_AUTH_RESPONSE = {
    "access_token": "123456789",
    "state": "somestate",
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import quote

import pytest
import responses

from tests.data.test_defaults import DEFAULT_OAUTH_URL
from tests.utils.test_utils import data_matcher, param_matcher
from todoist_api_python._core.endpoints import API_URL  # Use new base URL
from todoist_api_python.authentication import (
    get_auth_token,
    get_auth_token_async,
    get_authentication_url,
    revoke_auth_token,
    revoke_auth_token_async,
)

if TYPE_CHECKING:
    from todoist_api_python.models import AuthResult


def test_get_authentication_url() -> None:
    client_id = "123"
    scopes = ["task:add", "data:read", "project:delete"]
    state = "456"
    params = (
        f"client_id={client_id}&scope={scopes[0]},{scopes[1]},{scopes[2]}&state={state}"
    )
    query = quote(params, safe="=&")
    expected_url = f"{DEFAULT_OAUTH_URL}/authorize?{query}"

    url = get_authentication_url(client_id, scopes, state)

    assert url == expected_url


@pytest.mark.asyncio
async def test_get_auth_token(
    requests_mock: responses.RequestsMock,
    default_auth_response: dict[str, Any],
    default_auth_result: AuthResult,
) -> None:
    client_id = "123"
    client_secret = "456"
    code = "789"

    requests_mock.add(
        responses.POST,
        f"{DEFAULT_OAUTH_URL}/access_token",
        json=default_auth_response,
        status=200,
        match=[
            data_matcher(
                {"client_id": client_id, "client_secret": client_secret, "code": code}
            )
        ],
    )

    auth_result = get_auth_token(client_id, client_secret, code)

    assert len(requests_mock.calls) == 1
    assert auth_result == default_auth_result

    auth_result = await get_auth_token_async(client_id, client_secret, code)

    assert len(requests_mock.calls) == 2
    assert auth_result == default_auth_result


@pytest.mark.asyncio
async def test_revoke_auth_token(
    requests_mock: responses.RequestsMock,
) -> None:
    client_id = "123"
    client_secret = "456"
    token = "AToken"

    requests_mock.add(
        responses.DELETE,
        f"{API_URL}/access_tokens",
        match=[
            param_matcher(
                {
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "access_token": token,
                }
            )
        ],
        status=200,
    )

    result = revoke_auth_token(client_id, client_secret, token)

    assert len(requests_mock.calls) == 1
    assert result is True

    result = await revoke_auth_token_async(client_id, client_secret, token)

    assert len(requests_mock.calls) == 2
    assert result is True



from __future__ import annotations

from typing import Any, Literal
from urllib.parse import urlencode

import requests
from requests import Session

from todoist_api_python._core.endpoints import (
    ACCESS_TOKEN_PATH,
    ACCESS_TOKENS_PATH,
    AUTHORIZE_PATH,
    get_api_url,
    get_oauth_url,
)
from todoist_api_python._core.http_requests import delete, post
from todoist_api_python._core.utils import run_async
from todoist_api_python.models import AuthResult

"""
Possible permission scopes:

- `data:read`: Read-only access
- `data:read_write`: Read and write access
- `data:delete`: Full access including delete
- `task:add`: Can create new tasks
- `project:delete`: Can delete projects
- `backups:read`: Can access user backups without MFA
"""
Scope = Literal[
    "task:add",
    "data:read",
    "data:read_write",
    "data:delete",
    "project:delete",
    "backups:read",
]


def get_authentication_url(client_id: str, scopes: list[Scope], state: str) -> str:
    """Get authorization URL to initiate OAuth flow."""
    if len(scopes) == 0:
        raise ValueError("At least one authorization scope should be requested.")

    endpoint = get_oauth_url(AUTHORIZE_PATH)
    query = {
        "client_id": client_id,
        "scope": ",".join(scopes),
        "state": state,
    }
    return f"{endpoint}?{urlencode(query)}"


def get_auth_token(
    client_id: str, client_secret: str, code: str, session: Session | None = None
) -> AuthResult:
    """Get access token using provided client ID, client secret, and auth code."""
    endpoint = get_oauth_url(ACCESS_TOKEN_PATH)
    session = session or requests.Session()
    data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "code": code,
    }
    response: dict[str, Any] = post(session=session, url=endpoint, data=data)
    return AuthResult.from_dict(response)


async def get_auth_token_async(
    client_id: str, client_secret: str, code: str
) -> AuthResult:
    return await run_async(lambda: get_auth_token(client_id, client_secret, code))


def revoke_auth_token(
    client_id: str, client_secret: str, token: str, session: Session | None = None
) -> bool:
    """Revoke an access token."""
    # `get_api_url` is not a typo. Deleting access tokens is done using the regular API.
    endpoint = get_api_url(ACCESS_TOKENS_PATH)
    session = session or requests.Session()
    params = {
        "client_id": client_id,
        "client_secret": client_secret,
        "access_token": token,
    }
    return delete(session=session, url=endpoint, params=params)


async def revoke_auth_token_async(
    client_id: str, client_secret: str, token: str
) -> bool:
    return await run_async(lambda: revoke_auth_token(client_id, client_secret, token))



from __future__ import annotations

import sys
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    UTC = timezone.utc

import pytest
import responses

from tests.data.test_defaults import DEFAULT_API_URL, PaginatedItems
from tests.utils.test_utils import (
    auth_matcher,
    enumerate_async,
    param_matcher,
    request_id_matcher,
)
from todoist_api_python._core.utils import format_datetime

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import Task


@pytest.mark.asyncio
async def test_get_completed_tasks_by_due_date(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_completed_tasks_response: list[PaginatedItems],
    default_completed_tasks_list: list[list[Task]],
) -> None:
    since = datetime(2024, 1, 1, 0, 0, 0, tzinfo=UTC)
    until = datetime(2024, 2, 1, 0, 0, 0, tzinfo=UTC)
    project_id = "6X7rM8997g3RQmvh"
    filter_query = "p1"

    params = {
        "since": format_datetime(since),
        "until": format_datetime(until),
        "project_id": project_id,
        "filter_query": filter_query,
    }

    endpoint = f"{DEFAULT_API_URL}/tasks/completed/by_due_date"

    cursor: str | None = None
    for page in default_completed_tasks_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher(params, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    tasks_iter = todoist_api.get_completed_tasks_by_due_date(
        since=since,
        until=until,
        project_id=project_id,
        filter_query=filter_query,
    )

    for i, tasks in enumerate(tasks_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_completed_tasks_list[i]
        count += 1

    tasks_async_iter = await todoist_api_async.get_completed_tasks_by_due_date(
        since=since,
        until=until,
        project_id=project_id,
        filter_query=filter_query,
    )

    async for i, tasks in enumerate_async(tasks_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_completed_tasks_list[i]
        count += 1


@pytest.mark.asyncio
async def test_get_completed_tasks_by_completion_date(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_completed_tasks_response: list[PaginatedItems],
    default_completed_tasks_list: list[list[Task]],
) -> None:
    since = datetime(2024, 3, 1, 0, 0, 0)  # noqa: DTZ001
    until = datetime(2024, 4, 1, 0, 0, 0)  # noqa: DTZ001
    workspace_id = "123"
    filter_query = "@label"

    params: dict[str, Any] = {
        "since": format_datetime(since),
        "until": format_datetime(until),
        "workspace_id": workspace_id,
        "filter_query": filter_query,
    }

    endpoint = f"{DEFAULT_API_URL}/tasks/completed/by_completion_date"

    cursor: str | None = None
    for page in default_completed_tasks_response:
        requests_mock.add(
            method=responses.GET,
            url=endpoint,
            json=page,
            status=200,
            match=[auth_matcher(), request_id_matcher(), param_matcher(params, cursor)],
        )
        cursor = page["next_cursor"]

    count = 0

    tasks_iter = todoist_api.get_completed_tasks_by_completion_date(
        since=since,
        until=until,
        workspace_id=workspace_id,
        filter_query=filter_query,
    )

    for i, tasks in enumerate(tasks_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_completed_tasks_list[i]
        count += 1

    tasks_async_iter = await todoist_api_async.get_completed_tasks_by_completion_date(
        since=since,
        until=until,
        workspace_id=workspace_id,
        filter_query=filter_query,
    )

    async for i, tasks in enumerate_async(tasks_async_iter):
        assert len(requests_mock.calls) == count + 1
        assert tasks == default_completed_tasks_list[i]
        count += 1



from __future__ import annotations

import re
import unicodedata

API_VERSION = "v1"

API_URL = f"https://api.todoist.com/api/{API_VERSION}"
OAUTH_URL = "https://todoist.com/oauth"
PROJECT_URL = "https://app.todoist.com/app/project"
INBOX_URL = "https://app.todoist.com/app/inbox"
TASK_URL = "https://app.todoist.com/app/task"

TASKS_PATH = "tasks"
TASKS_FILTER_PATH = "tasks/filter"
TASKS_QUICK_ADD_PATH = "tasks/quick"
TASKS_COMPLETED_PATH = "tasks/completed"
TASKS_COMPLETED_BY_DUE_DATE_PATH = f"{TASKS_COMPLETED_PATH}/by_due_date"
TASKS_COMPLETED_BY_COMPLETION_DATE_PATH = f"{TASKS_COMPLETED_PATH}/by_completion_date"
PROJECTS_PATH = "projects"
PROJECTS_SEARCH_PATH_SUFFIX = "search"
PROJECT_ARCHIVE_PATH_SUFFIX = "archive"
PROJECT_UNARCHIVE_PATH_SUFFIX = "unarchive"
COLLABORATORS_PATH = "collaborators"
SECTIONS_PATH = "sections"
SECTIONS_SEARCH_PATH_SUFFIX = "search"
COMMENTS_PATH = "comments"
LABELS_PATH = "labels"
LABELS_SEARCH_PATH_SUFFIX = "search"
SHARED_LABELS_PATH = "labels/shared"
SHARED_LABELS_RENAME_PATH = f"{SHARED_LABELS_PATH}/rename"
SHARED_LABELS_REMOVE_PATH = f"{SHARED_LABELS_PATH}/remove"

AUTHORIZE_PATH = "authorize"
ACCESS_TOKEN_PATH = "access_token"  # noqa: S105
ACCESS_TOKENS_PATH = "access_tokens"


def get_oauth_url(relative_path: str) -> str:
    """
    Generate the URL for a given OAuth endpoint.

    :param relative_path: The relative path of the endpoint.
    :return: The URL string for the OAuth endpoint.
    """
    return f"{OAUTH_URL}/{relative_path}"


def get_api_url(relative_path: str) -> str:
    """
    Generate the URL for a given API endpoint.

    :param relative_path: The relative path of the endpoint.
    :return: The URL string for the API endpoint.
    """
    return f"{API_URL}/{relative_path}"


def get_task_url(task_id: str, content: str | None = None) -> str:
    """
    Generate the URL for a given task.

    :param task_id: The ID of the task.
    :param content: The content of the task.
    :return: The URL string for the task view.
    """
    slug = _slugify(content) if content is not None else None
    path = f"{slug}-{task_id}" if content else task_id
    return f"{TASK_URL}/{path}"


def get_project_url(project_id: str, name: str | None = None) -> str:
    """
    Generate the URL for a given project.

    :param project_id: The ID of the project.
    :param name: The name of the project.
    :return: The URL string for the project view.
    """
    slug = _slugify(name) if name is not None else None
    path = f"{slug}-{project_id}" if name else project_id
    return f"{PROJECT_URL}/{path}"


def _slugify(value: str) -> str:
    """
    Slugify function borrowed from Django.

    Convert to ASCII. Convert spaces or repeated dashes to single dashes.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Strip spaces, dashes, and underscores.
    """
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



from __future__ import annotations

from tests.data.test_defaults import (
    DEFAULT_ATTACHMENT_RESPONSE,
    DEFAULT_COLLABORATOR_RESPONSE,
    DEFAULT_COMMENT_RESPONSE,
    DEFAULT_DUE_RESPONSE,
    DEFAULT_DURATION_RESPONSE,
    DEFAULT_LABEL_RESPONSE,
    DEFAULT_PROJECT_RESPONSE,
    DEFAULT_PROJECT_RESPONSE_2,
    DEFAULT_SECTION_RESPONSE,
    DEFAULT_TASK_RESPONSE,
)
from todoist_api_python._core.utils import parse_date, parse_datetime
from todoist_api_python.models import (
    Attachment,
    AuthResult,
    Collaborator,
    Comment,
    Due,
    Duration,
    Label,
    Project,
    Section,
    Task,
)

unexpected_data = {"unexpected_key": "some value"}


def test_due_from_dict() -> None:
    sample_data = dict(DEFAULT_DUE_RESPONSE)
    sample_data.update(unexpected_data)

    due = Due.from_dict(sample_data)

    assert due.date == parse_date(str(sample_data["date"]))
    assert due.timezone == sample_data["timezone"]
    assert due.string == sample_data["string"]
    assert due.lang == sample_data["lang"]
    assert due.is_recurring == sample_data["is_recurring"]


def test_duration_from_dict() -> None:
    sample_data = dict(DEFAULT_DURATION_RESPONSE)
    sample_data.update(unexpected_data)

    duration = Duration.from_dict(sample_data)

    assert duration.amount == sample_data["amount"]
    assert duration.unit == sample_data["unit"]


def test_project_from_dict() -> None:
    sample_data = dict(DEFAULT_PROJECT_RESPONSE)
    sample_data.update(unexpected_data)

    project = Project.from_dict(sample_data)

    assert project.id == sample_data["id"]
    assert project.name == sample_data["name"]
    assert project.description == sample_data["description"]
    assert project.parent_id == sample_data["parent_id"]
    assert project.folder_id == sample_data["folder_id"]
    assert project.workspace_id == sample_data["workspace_id"]
    assert project.order == sample_data["child_order"]
    assert project.color == sample_data["color"]
    assert project.is_collapsed == sample_data["collapsed"]
    assert project.is_shared == sample_data["shared"]
    assert project.is_favorite == sample_data["is_favorite"]
    assert project.is_inbox_project == sample_data["is_inbox_project"]
    assert project.can_assign_tasks == sample_data["can_assign_tasks"]
    assert project.view_style == sample_data["view_style"]
    assert project.created_at == parse_datetime(str(sample_data["created_at"]))
    assert project.updated_at == parse_datetime(str(sample_data["updated_at"]))


def test_project_url() -> None:
    inbox = Project.from_dict(dict(DEFAULT_PROJECT_RESPONSE))
    assert inbox.url == "https://app.todoist.com/app/inbox"
    project = Project.from_dict(dict(DEFAULT_PROJECT_RESPONSE_2))
    assert project.url == "https://app.todoist.com/app/project/inbox-6X7rfFVPjhvv84XG"


def test_task_from_dict() -> None:
    sample_data = dict(DEFAULT_TASK_RESPONSE)
    sample_data.update(unexpected_data)

    task = Task.from_dict(sample_data)

    assert task.id == sample_data["id"]
    assert task.content == sample_data["content"]
    assert task.description == sample_data["description"]
    assert task.project_id == sample_data["project_id"]
    assert task.section_id == sample_data["section_id"]
    assert task.parent_id == sample_data["parent_id"]
    assert task.labels == sample_data["labels"]
    assert task.priority == sample_data["priority"]
    assert task.due == Due.from_dict(sample_data["due"])
    assert task.duration == Duration.from_dict(sample_data["duration"])
    assert task.is_collapsed == sample_data["collapsed"]
    assert task.order == sample_data["child_order"]
    assert task.assignee_id == sample_data["responsible_uid"]
    assert task.assigner_id == sample_data["assigned_by_uid"]
    assert task.completed_at == sample_data["completed_at"]
    assert task.creator_id == sample_data["added_by_uid"]
    assert task.created_at == parse_datetime(sample_data["added_at"])
    assert task.updated_at == parse_datetime(sample_data["updated_at"])


def test_task_url() -> None:
    task = Task.from_dict(dict(DEFAULT_TASK_RESPONSE))
    assert (
        task.url
        == "https://app.todoist.com/app/task/some-task-content-6X7rM8997g3RQmvh"
    )


def test_section_from_dict() -> None:
    sample_data = dict(DEFAULT_SECTION_RESPONSE)
    sample_data.update(unexpected_data)

    section = Section.from_dict(sample_data)

    assert section.id == sample_data["id"]
    assert section.project_id == sample_data["project_id"]
    assert section.name == sample_data["name"]
    assert section.order == sample_data["order"]


def test_collaborator_from_dict() -> None:
    sample_data = dict(DEFAULT_COLLABORATOR_RESPONSE)
    sample_data.update(unexpected_data)

    collaborator = Collaborator.from_dict(sample_data)

    assert collaborator.id == sample_data["id"]
    assert collaborator.email == sample_data["email"]
    assert collaborator.name == sample_data["name"]


def test_attachment_from_dict() -> None:
    sample_data = dict(DEFAULT_ATTACHMENT_RESPONSE)
    sample_data.update(unexpected_data)

    attachment = Attachment.from_dict(sample_data)

    assert attachment.resource_type == sample_data["resource_type"]
    assert attachment.file_name == sample_data["file_name"]
    assert attachment.file_size == sample_data["file_size"]
    assert attachment.file_type == sample_data["file_type"]
    assert attachment.file_url == sample_data["file_url"]
    assert attachment.upload_state == sample_data["upload_state"]
    assert attachment.image == sample_data["image"]
    assert attachment.image_width == sample_data["image_width"]
    assert attachment.image_height == sample_data["image_height"]
    assert attachment.url == sample_data["url"]
    assert attachment.title == sample_data["title"]


def test_comment_from_dict() -> None:
    sample_data = dict(DEFAULT_COMMENT_RESPONSE)
    sample_data.update(unexpected_data)

    comment = Comment.from_dict(sample_data)

    assert comment.id == sample_data["id"]
    assert comment.content == sample_data["content"]
    assert comment.poster_id == sample_data["posted_uid"]
    assert comment.posted_at == parse_datetime(sample_data["posted_at"])
    assert comment.task_id == sample_data["task_id"]
    assert comment.project_id == sample_data["project_id"]
    assert comment.attachment == Attachment.from_dict(sample_data["attachment"])


def test_label_from_dict() -> None:
    sample_data = dict(DEFAULT_LABEL_RESPONSE)
    sample_data.update(unexpected_data)

    label = Label.from_dict(sample_data)

    assert label.id == sample_data["id"]
    assert label.name == sample_data["name"]
    assert label.color == sample_data["color"]
    assert label.order == sample_data["order"]
    assert label.is_favorite == sample_data["is_favorite"]


def test_auth_result_from_dict() -> None:
    token = "123"
    state = "456"
    sample_data = {"access_token": token, "state": state}
    sample_data.update(unexpected_data)

    auth_result = AuthResult.from_dict(sample_data)

    assert auth_result.access_token == token
    assert auth_result.state == state
